unit xwindowlist;

{$MODE objfpc}{$H+}

interface

uses
  Classes, SysUtils, x, xlib, xutil, unixtype, fgl;

type

  TCardinalList = specialize TFPGList<cardinal>;
  TWindowProc = procedure(Win: TWindow) of object;

  { TXWindowList }

  TXWindowList = class
  private
    fList: TCardinalList;
    fDisplay: PDisplay;
    fNet_Wm_Pid: TAtom;
    fWm_Client_machine: TAtom;
    procedure EnumerateTopLevelWindows(Proc: TWindowProc; Display: PDisplay);
    function GetWindowList: TCardinalList;
    procedure XWindowProc(Win: TWindow);
  public
    constructor Create;
    destructor Destroy; override;
    procedure UpdateWindowList;
    property WindowList: TCardinalList read GetWindowList;
    function GetWindowName(Window: TWindow): string;
    function GetWindowPID(Window: TWindow): cardinal;
    function GetWindowState(Window: TWindow): string;
    function GetWindowCmd(PID: cardinal): string;
    function GetWindowHost(Window: TWindow): string;
    function GetWindowRect(Window: TWindow): TRect;
    procedure ActivateWindow(Window: TWindow);
    procedure ResizeWindow(Window: TWindow; Width, Height: integer);
  end;

  TWindowArray = array[0..MaxListSize] of TWindow;
  PWindowArray = ^TWindowArray;

function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: cardinal;
  Offset: cardinal = 0): boolean; overload;
function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: string;
  Offset: cardinal = 0): boolean; overload;
function GetState(Dpy: PDisplay; Win: TWindow): cardinal;
function GetDesktop(Dpy: PDisplay; Win: TWindow): cardinal;
function SetDesktop(Dpy: PDisplay; Desktop: cardinal): boolean;

threadvar
  GotBadwindow: boolean;

threadvar
  OldHandler: TXErrorHandler;


implementation

function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: cardinal;
  Offset: cardinal): boolean;
var
  ActualTypeReturn: TAtom;
  ActualFormatReturn: longint;
  NItemsReturn, BytesAfterReturn: cardinal;
  Ptr: pbyte;
begin
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 1, longbool(0),
    AnyPropertyType, @ActualTypeReturn, @ActualFormatReturn, @NItemsReturn,
    @BytesAfterReturn, @Ptr) = Success;
  if Result then
    try
    {$message warn: ActualFormatReturn always 0 on 64bit}
      if (ActualTypeReturn = None) {or (ActualFormatReturn <> 32) } or
        not Assigned(Ptr) then
      begin
        Result := False;
      end;
      if Result then Value := PCardinal(Ptr)^;
    finally
      if Assigned(Ptr) then XFree(Ptr);
    end;
end;

function GetProp(Dpy: PDisplay; Win: TWindow; Prop: TAtom; out Value: string;
  Offset: cardinal = 0): boolean;
var
  ActualTypeReturn: TAtom;
  ActualFormatReturn: longint;
  Length, NItemsReturn, BytesAfterReturn: cardinal;
  Ptr: pbyte;
begin
  Result := XGetWindowProperty(Dpy, Win, Prop, Offset, 0, longbool(0),
    None, @ActualTypeReturn, @ActualFormatReturn, @NItemsReturn,
    @BytesAfterReturn, @Ptr) = Success;
  if Result and Assigned(Ptr) then
    XFree(Ptr);
  if (ActualTypeReturn = None) or (ActualFormatReturn <> 8) then Result := False;
  if Result then
    try
      Length := (BytesAfterReturn + 3) div 4;
      Result := XGetWindowProperty(Dpy, Win, Prop, Offset, Length,
        longbool(0), ActualTypeReturn, @ActualTypeReturn, @ActualFormatReturn,
        @NItemsReturn, @BytesAfterReturn, @Ptr) = Success;
      if (ActualTypeReturn = None) or (ActualFormatReturn <> 8) or not Assigned(Ptr) then
        Result := False;
      if Result then
        SetString(Value, PChar(Ptr), NItemsReturn);
    finally
      if Assigned(Ptr) then
        XFree(Ptr);
    end;
end;



function GetState(Dpy: PDisplay; Win: TWindow): cardinal;
var
  A: TAtom;
begin
  A := XInternAtom(Dpy, 'WM_STATE', longbool(1));
  if A = None then
    Result := 0
  else if not GetProp(Dpy, Win, A, Result) then
    Result := 0;
end;

function GetDesktop(Dpy: PDisplay; Win: TWindow): cardinal;
var
  A: TAtom;
begin
  A := XInternAtom(Dpy, '_NET_WM_DESKTOP', longbool(1));
  if A = None then
    Result := $FFFFFFFF
  else if not GetProp(Dpy, Win, A, Result) then
    Result := $FFFFFFFF;
end;

function SetDesktop(Dpy: PDisplay; Desktop: cardinal): boolean;
var
  Net_Current_Desktop: TAtom;
  Root: TWindow;
  CurrentDesktop: cardinal;
  Event: TXEvent;
begin
  Net_Current_Desktop := XInternAtom(Dpy, '_NET_CURRENT_DESKTOP', longbool(1));
  if Net_Current_Desktop = None then
    Result := False
  else
  begin
    Root := XDefaultRootWindow(Dpy);
    if not GetProp(Dpy, Root, Net_Current_Desktop, CurrentDesktop) then
      Result := False
    else
    begin
      Event.xclient._type := ClientMessage;
      Event.xclient.send_event := Tbool(True);
      Event.xclient.window := Root;
      Event.xclient.display := Dpy;
      Event.xclient.format := 32;
      Event.xclient.message_type := Net_Current_Desktop;
      Event.xclient.Data.l[0] := Desktop;
      Result := XSendEvent(Dpy, Root, longbool(0), SubstructureNotifyMask or
        SubstructureRedirectMask, @Event) = Success;
    end;
  end;
end;


function BadwindowHandler(para1: PDisplay; para2: PXErrorEvent): cint; cdecl;
begin
  if para2^.error_code <> BadWindow then
    Result := OldHandler(para1, para2)
  else
  begin
    GotBadwindow := True;
    Result := 0;
  end;
end;

procedure TXWindowList.EnumerateTopLevelWindows(Proc: TWindowProc; Display: PDisplay);

  procedure QueryChildren(Dpy: PDisplay; Win: TWindow);
  var
    Root, Parent: TWindow;
    Children: PWindowArray;
    I, Count: cardinal;
  begin
    if XQueryTree(Dpy, Win, @Root, @Parent, PPWindow(@Children), @Count) = 0 then Exit;
    if Count = 0 then Exit;
    try
      for I := 0 to Count - 1 do
      begin
        GotBadwindow := False;
        QueryChildren(Dpy, Children^[I]);
        XSync(dpy, False);
        if (not GotBadwindow) and (GetState(Dpy, Children^[I]) <> 0) then
        begin
          Proc(Children^[I]);
        end;
      end;
    finally
      XFree(Children);
    end;
  end;

var
  Root: TWindow;

begin
  try
    Root := XRootWindow(Display, XDefaultScreen(Display));
    OldHandler := XSetErrorHandler(@BadwindowHandler);
    QueryChildren(Display, Root);
  finally
    XSetErrorHandler(OldHandler);
  end;
end;


{ TXWindowList }

function TXWindowList.GetWindowList: TCardinalList;
begin
  Result := fList;
end;

procedure TXWindowList.XWindowProc(Win: TWindow);
begin
  fList.Add(Win);
end;

constructor TXWindowList.Create;
begin
  fDisplay := XOpenDisplay(PChar(GetEnvironmentVariable('DISPLAY')));
  if not Assigned(fDisplay) then raise Exception.Create('Can not open display');
  fList := TCardinalList.Create;
  fWm_Client_machine := XInternAtom(fDisplay, 'WM_CLIENT_MACHINE', longbool(1));
  fNet_Wm_Pid := XInternAtom(fDisplay, '_NET_WM_PID', longbool(1));
end;

destructor TXWindowList.Destroy;
begin
  fList.Free;
  if Assigned(fDisplay) then XCloseDisplay(fDisplay);
  inherited Destroy;
end;

procedure TXWindowList.UpdateWindowList;
begin
  fList.Clear;
  EnumerateTopLevelWindows(@XWindowProc, fDisplay);
end;

function TXWindowList.GetWindowName(Window: TWindow): string;
var
  Stat: XLib.TStatus;
  text_prop: TXTextProperty;
  List: PPChar;
  Num: integer;
begin
  Result := '';
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  stat := XGetWMName(fDisplay, Window, @text_prop);
  XSync(fDisplay, False);
  if GotBadwindow then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end;
  if ((stat = 0) or (text_prop.Value = nil) or (text_prop.nitems = 0)) then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end;
  stat := XmbTextPropertyToTextList(fDisplay, @text_prop, @List, @Num);
  if ((stat < Success) or (num = 0) or (list = nil)) then
  begin
    XSetErrorHandler(OldHandler);
    exit;
  end
  else
  begin
    XFree(text_prop.Value);
    if Num > 0 then Result := List^;
    XFreeStringList(list);
  end;
  XSetErrorHandler(OldHandler);
end;

function TXWindowList.GetWindowPID(Window: TWindow): cardinal;
var
  Pid: cardinal;
begin
  Result := 0;
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  if GetProp(fDisplay, Window, fNet_Wm_Pid, Pid) then
    if not GotBadwindow then Result := Pid;
  XSetErrorHandler(OldHandler);
end;

function StateName(State: cardinal): string;
begin
  case State of
    IconicState: Result := 'Iconic';
    NormalState: Result := 'Normal';
    WithdrawnState: Result := 'Withdrawn';
    else
      Result := IntToStr(State);
  end;
end;


function TXWindowList.GetWindowState(Window: TWindow): string;
var
  stat: cardinal;
begin
  Result := '';
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  Stat := GetState(fDisplay, Window);
  XSync(fDisplay, False);
  if not GotBadwindow then Result := StateName(Stat);
  XSetErrorHandler(OldHandler);
end;

function TXWindowList.GetWindowCmd(PID: cardinal): string;
var
  sl: TStringList;
begin
  Result := '';
  sl := TStringList.Create;
  try
    if FileExists('/proc/' + IntToStr(Pid) + '/cmdline') then
    begin
      sl.LoadFromFile('/proc/' + IntToStr(Pid) + '/cmdline');
      if sl.Count > 0 then
      begin
        sl.Text := StringReplace(sl.Text, #0, LineEnding, [rfReplaceAll]);
        Result := sl[0];
      end;
    end;
  finally
    sl.Free;
  end;
end;

function TXWindowList.GetWindowHost(Window: TWindow): string;
var
  Host: string;
begin
  Result := '';
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  if GetProp(fDisplay, Window, fWm_Client_Machine, Host) then
    if not GotBadwindow then Result := Host;
  XSetErrorHandler(OldHandler);
end;

function TXWindowList.GetWindowRect(Window: TWindow): TRect;
var
  Stat: TStatus;
  Attrs: TXWindowAttributes;
  rx, ry: cint;
  junk: TWindow;
begin
  Result := Rect(0, 0, 0, 0);
  OldHandler := XSetErrorHandler(@BadwindowHandler);
  GotBadwindow := False;
  Stat := XGetWindowAttributes(fDisplay, Window, @Attrs);
  XSync(fDisplay, False);
  if (not GotBadwindow) and (Stat <> 0) then
  begin
    XTranslateCoordinates(fDisplay, Window, Attrs.root, -Attrs.border_width,
      -Attrs.border_width, @rx, @ry, @junk);
    Result := Rect(rx, ry, rx + Attrs.Width, ry + Attrs.Height);
  end;
  XSetErrorHandler(OldHandler);
end;

procedure TXWindowList.ResizeWindow(Window: TWindow; Width, Height:integer);
begin
    XResizeWindow(FDisplay, Window, Width, Height);
    XFlush(Fdisplay);
end;

procedure TXWindowList.ActivateWindow(Window: TWindow);
var
  Desktop: cardinal;
  atom: TAtom;
  xev: TXEvent;
  attr: TXWindowAttributes;
begin
  Desktop := GetDesktop(fDisplay, Window);
  if Desktop <> $FFFFFFFF then
  begin
    SetDesktop(fDisplay, Desktop);
    XGetWindowAttributes(fDisplay, Window, @attr);

    if (attr.map_state = IsViewable) or ((attr.map_state = IsUnmapped) and
      (Getstate(fDisplay, Window) = IconicState)) then
    begin
      XMapWindow(fDisplay, Window);
      XFlush(fDisplay);
      XSync(fDisplay, False);
    end;
    sleep(100); //Naja, wie wartet man darauf?
    //while (XPending(fDisplay)=1) do XNextEvent(fDisplay, @xev);

    XCirculateSubwindowsUp(fDisplay, Window);
    XRaiseWindow(fDisplay, Window);
    XMapWindow(fDisplay, Window);
    XSetInputFocus(fDisplay, Window, RevertToParent, CurrentTime);

    atom := XInternAtom(fDisplay, '_NET_ACTIVE_WINDOW', False);

    xev.xclient._type := ClientMessage;
    xev.xclient.serial := 0;
    xev.xclient.send_event := 1;
    xev.xclient.display := fDisplay;
    xev.xclient.window := Window;
    xev.xclient.message_type := atom;
    xev.xclient.format := 32;
    xev.xclient.Data.l[0] := 2;
    xev.xclient.Data.l[1] := 0;
    xev.xclient.Data.l[2] := 0;
    xev.xclient.Data.l[3] := 0;
    xev.xclient.Data.l[4] := 0;

    XSendEvent(fDisplay,
      attr.root, False,
      SubstructureRedirectMask or SubstructureNotifyMask, @xev);

    XFlush(fDisplay);
    XSync(fDisplay, False);
  end;
end;

end.
